#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name="raxterize",
    version="0.1.0",
    description="Razterization of 2D meshes using JAX",
    author="Magnus Dalen Kvalevåg",
    author_email="magnus.kvalevag@ntnu.no",
    packages=find_packages(),
)
